<?php

namespace App\domain\travelport;

use App\Jobs\TravelportCombinedOffersSavingJob;
use App\Jobs\TravelportReferenceListSavingJob;
use Carbon\Carbon;
use Carbon\CarbonInterval;
use Exception;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Log;

class QueryCombinableFlightOffersAction
{
    protected string $catalogProductOfferingId;
    protected Collection $referenceList;

    protected string $requestId;

    protected FlightOfferService $flightOfferService;

    public function __construct()
    {
        $this->flightOfferService = new FlightOfferService();
        $this->referenceList = new Collection;
    }

    public function execute(array $passengerCriteria, array $searchCriteriaFlight, ?array $searchModifiers): array
    {
        $this->requestId = uuid_create();

        Log::info("### Request ID: " . $this->requestId);

        $numberOfLegs = count($searchCriteriaFlight);

        $combinedOffers = $this->getCombinedOffersFromAllSource(
            $passengerCriteria,
            $searchCriteriaFlight,
            $searchModifiers
        );

        $sortedCombinedOffers = $this->sortAndRemovedUnwantedCombinedOffers($combinedOffers, $numberOfLegs);

        dispatch(new TravelportCombinedOffersSavingJob($this->requestId, $sortedCombinedOffers));

        session(['total_combined_offers' => $sortedCombinedOffers->count()]);
        session(['offers_request_id' => $this->requestId]);

        return $this->attachReferenceDataOnlyForInitialPage($sortedCombinedOffers)->toArray();
    }

    protected function getCombinedOffersFromAllSource(
        array  $passengerCriteria,
        array  $searchCriteriaFlight,
        ?array $searchModifiers
    ): Collection
    {
        $sources = ["NDC", "GDS"];
        $totalCombinedOffers = new Collection;

        foreach ($sources as $source) {
            $totalCombinedOffers = $totalCombinedOffers->merge(
                $this->getCombinableOffers($source, $passengerCriteria, $searchCriteriaFlight, $searchModifiers)
            );
        }

        return $totalCombinedOffers;
    }

    protected function getCombinableOffers(
        string $source,
        array  $passengerCriteria,
        array  $searchCriteriaFlight,
        ?array $searchModifiers
    ): Collection
    {
        $offersResponse = $this->flightOfferService->requestOffers(
            $passengerCriteria,
            $searchCriteriaFlight,
            $searchModifiers,
            $source
        );

//        session([$source => $offersResponse]);

        Log::info('Content received..');

        if ($offersResponse) {
//            Log::info($source . ' contents' . json_encode($offersResponse));
            return $this->generateCombinableOffers($source, $offersResponse);
        }

        return new Collection([]);
    }

    protected function sortAndRemovedUnwantedCombinedOffers(Collection $totalCombinedOffers, int $numberOfLegs): Collection
    {
        $sortedCombinedOffers = $totalCombinedOffers
            ->sortBy(fn($combination) => $combination[0]['bestCombinablePrice']['TotalPrice']);

        if ($numberOfLegs > 1) {
            $sortedCombinedOffers = $sortedCombinedOffers
                ->filter(fn($combinableOffer) => count($combinableOffer) > 1);
        }
        return $sortedCombinedOffers;
    }

    protected function generateCombinableOffers(string $source, array $offersContents): Collection
    {
        $catalogProductOfferings = new Collection(
            collect($offersContents['CatalogProductOfferingsResponse']['CatalogProductOfferings'])
        );
        $catalogProductOffering = new Collection($catalogProductOfferings['CatalogProductOffering']);

        $this->referenceList->add([
            "source" => $source,
            "catalogId" => $catalogProductOfferings['Identifier']['value'],
            "content" => new Collection($offersContents['CatalogProductOfferingsResponse']['ReferenceList'])
        ]);

        session(["$source'_catalog_id'" => $catalogProductOfferings['Identifier']['value']]);

        Log::info("$source'_catalog_id'");
        Log::info('session value: ' . session("$source'_catalog_id'"));

        dispatch(new TravelportReferenceListSavingJob($this->requestId, $this->referenceList));

        $productOffers = $catalogProductOffering->flatMap(function ($offer) {
            if (!isset($offer['ProductBrandOptions'])) {
                return [];
            }

            return collect($offer['ProductBrandOptions'])->flatMap(
                function ($productOption) use ($offer) {

                    return $this->generateBrandOptions($offer, $productOption);
                });
        });

        Log::info("Product offers generated.");
        return $productOffers->groupBy('combinabilityCode');
    }

    protected function generateBrandOptions($offer, $productOption): Collection
    {
        return collect($productOption['ProductBrandOffering'])->flatMap(
            function ($productBrandOffering) use ($offer) {

                return $this->generateProductBrands($productBrandOffering, $offer);
            });
    }

    protected function generateProductBrands($productBrandOffering, $offer): Collection
    {
        return collect($productBrandOffering['CombinabilityCode'])->map(
            function ($combinabilityCode) use ($offer, $productBrandOffering) {

                return $this->generateOfferProduct($combinabilityCode, $offer, $productBrandOffering);
            });
    }

    protected function generateOfferProduct($combinabilityCode, $offer, $productBrandOffering): array
    {
        $contentSource = (isset($offer['Identifier']) && $offer['Identifier']['authority']) ? "NDC" : "GDS";
        return [
            '@type' => intval($offer['sequence']) === 1 ? 'Outbound' : 'Inbound',
            'combinabilityCode' => $combinabilityCode,
            'catalogProductOfferingsIdentifier' => session("$contentSource'_catalog_id'"),
            'authority' => $offer['Identifier']['authority'] ?? "GDS",
            'offerId' => $offer['id'],
            'id' => $productBrandOffering['Product'][0]['productRef'],
            'fromAirport' => $offer['Departure'],
            'toAirport' => $offer['Arrival'],
            'numberOfBookableSeats' => $productIAir['Quantity'] ?? 0,
            'price' => $productBrandOffering['Price'],
            'bestCombinablePrice' => $productBrandOffering['BestCombinablePrice'],
        ];
    }

    protected function calculateDirectionDuration(Collection $segments): array
    {
        try {
            $durations = $segments->pluck('FlightDetails.duration');

            $totalHours = 0;
            $totalMinutes = 0;

            $durations->each(function ($durationString) use (&$totalHours, &$totalMinutes) {
                $duration = CarbonInterval::fromString($durationString);
                $totalHours += $duration->hours;
                $totalMinutes += $duration->minutes;
            });

            $totalHours += floor($totalMinutes / 60);
            $totalMinutes = $totalMinutes % 60;

            $durationHours = $totalHours + ($totalMinutes ? $totalMinutes / 60 : 0);

            return ["Duration" => "PT{$totalHours}H{$totalMinutes}M", "Hours" => $durationHours];

        } catch (Exception $e) {
            return ["Duration" => "", "Hours" => ""];
        }
    }

    private function getDepartureOrArrivalTime(Collection $firstFlight, string $direction): string
    {
        $departureDate = $firstFlight->get($direction)['date'];
        $departureTime = $firstFlight->get($direction)['time'];

        return Carbon::createFromFormat('Y-m-d H:i:s', $departureDate . ' ' . $departureTime)->toISOString();
    }

    protected function attachReferenceDataOnlyForInitialPage(Collection $totalCombinableOffers): Collection
    {
        return $totalCombinableOffers->take(20)->map(function (Collection $combinableOffers) {
            return $combinableOffers->map(function ($offer) {
                $productIAir = $this->extractProductAirFromReferenceList($offer['authority'], $offer['id']);
                $flightSegments = new Collection($productIAir['FlightSegment']);
                $flightSegments = $this->addFlightDetailsToSegments($offer['authority'], $flightSegments);
                $segmentDetails = $this->extractFlightSegmentDetails($flightSegments);
                $durations = $this->calculateProductDurations($productIAir);

                return [
                    ...$offer,
                    'productAir' => $productIAir,
                    'fromAirportCarrierCode' => $segmentDetails['fromAirportCarrierCode'],
                    'fromAirportAircraftCode' => $segmentDetails['fromAirportAircraftCode'],
                    'toAirportCarrierCode' => $segmentDetails['toAirportCarrierCode'],
                    'toAirportAircraftCode' => $segmentDetails['toAirportAircraftCode'],
                    'originDepartureTime' => $segmentDetails['originDepartureTime'],
                    'destinationArrivalTime' => $segmentDetails['destinationArrivalTime'],
                    'duration' => $durations['Duration'],
                    'numberOfStops' => $flightSegments->count() - 1,
                    'segments' => $flightSegments->toArray(),
                    'durationInHours' => $durations['Hours'],
                ];
            });
        });
    }

    protected function extractProductAirFromReferenceList(string $authority, string $productRef)
    {
        $referenceList = $this->getReferenceListOfGiveAuthority($authority);

        $productAirList = new Collection(
            $referenceList->where("@type", "ReferenceListProduct")->first()["Product"]
        );

        return $productAirList->where("id", $productRef)->first();
    }

    protected function addFlightDetailsToSegments(string $authority, Collection $flightSegments): Collection
    {
        $referenceList = $this->getReferenceListOfGiveAuthority($authority);

        $referenceFlights = new Collection(
            $referenceList->where("@type", "ReferenceListFlight")->first()['Flight']
        );

        return $flightSegments->map(function ($segment) use ($referenceFlights) {
            $flightDetails = $referenceFlights->where('id', $segment['Flight']['FlightRef'])->first();

            $segment['FlightDetails'] = $flightDetails;

            return $segment;
        });
    }

    protected function extractFlightSegmentDetails(Collection $segments): array
    {
        $duration = $this->calculateDirectionDuration($segments);

        $firstFlight = Collection::wrap($segments->first()["FlightDetails"]);
        $lastFlight = Collection::wrap($segments->last()["FlightDetails"]);
        return [
            'originDepartureTime' => $this->getDepartureOrArrivalTime($firstFlight, direction: "Departure"),
            'destinationArrivalTime' => $this->getDepartureOrArrivalTime($lastFlight, direction: "Arrival"),
            'duration' => $duration['Duration'],
            'durationInHours' => $duration['Hours'],
            'fromAirportCarrierCode' => $firstFlight['carrier'],
            'toAirportCarrierCode' => $lastFlight['carrier'],
            'fromAirportAircraftCode' => $firstFlight['equipment'],
            'toAirportAircraftCode' => $lastFlight['equipment'],
            'fromAirline' => isset($lastFlight['operatingCarrierName'])
                ? $lastFlight['operatingCarrierName']
                : "",
        ];
    }

    protected function calculateProductDurations(array $productAir): array
    {
        try {
            $totalHours = 0;
            $totalMinutes = 0;

            $totalDuration = CarbonInterval::fromString($productAir["totalDuration"]);

            $totalHours += $totalDuration->hours;
            $totalMinutes += $totalDuration->minutes;

            $totalHours += floor($totalMinutes / 60);
            $totalMinutes = $totalMinutes % 60;

            $durationHours = $totalHours + ($totalMinutes ? $totalMinutes / 60 : 0);

            return ["Duration" => $productAir["totalDuration"], "Hours" => $durationHours];

        } catch (Exception $e) {
            return ["Duration" => "", "Hours" => ""];
        }
    }

    protected function getReferenceListOfGiveAuthority(string $authority): mixed
    {
        $source = $authority === "GDS" ? "GDS" : "NDC";
        return $this->referenceList->where('source', $source)->first()['content'];
    }
}
